#-----------------------------------------------------------
# LibNac.pm
#
# Main Author: Hao Bryan Cheng (hbcheng@berkeley.edu)
#
# Holds common functions used by the circe perl scripts.
#
# Tested with Perl 5.10.1, 5.8.9
#-----------------------------------------------------------

package LibNAC;

#use feature qw{ switch }; # Can't use anymore for 5.8.9 compatibility
use strict;
use warnings;

use Sys::Syslog qw( :DEFAULT setlogsock);
use Net::IP::Match::XS;
use TryCatch;
use Data::Validate::IP qw(is_ipv4 is_ipv6);
use IO::Socket::INET;
use Crypt::CBC;
use constant {
	ACK	=> 0,
    ERR_INVALID_NETMASK => 2,
    ERR_INVALID_ROUTER => 3,
    ERR_ROUTER_NOT_IN_SUBNET => 4,
    ERR_INVALID_IP_IN_RANGE => 5,
    ERR_MALFORMED_CLIENT_RANGE => 6,
    ERR_NO_MATCHING_IFACE => 7,
	ERR_MISSING_CONFIG => 8,
	ERR_INVALID_CONFIG => 9,
	ERR_INVALID_NET_COMPRESSION => 10,
	ERR_INVALID_CONTEXT => 11,
	ERR_CANNOT_UNLOCK => 12,
	ERR_NO_CONTEXT_TABLE => 13,
	ERR_PUB_IP_IN_USE => 14,
	ERR_PRIV_IP_IN_USE => 15,
	ERR_CANNOT_DEL_BRIDGE => 16,
	ERR_INVALID_IP => 17,
	ERR_NAT_FAILURE => 21,
	ERR_IPFW_FAILURE => 22,
	ERR_ARP_FAILURE => 23,
	ERR_IFCONFIG_FAILURE => 24,
	ERR_INVALID_SYNTAX => 100,
	ACK_REQUEUE => 101,
	NAK_MALFORMED => 111,
	NAK_DUPLICATE => 121,
	NAK_INVALID => 122,
	NAK_STALE => 123,
};

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(nac_readconf circe_set_ipfw_context circe_set_context nac_flush_state nac_redirect);

our %EXPORT_TAGS = (constants => [qw(ERR_INVALID_NETMASK ERR_INVALID_ROUTER ERR_ROUTER_NOT_IN_SUBNET
ERR_INVALID_IP_IN_RANGE ERR_MALFORMED_CLIENT_RANGE ERR_NO_MATCHING_IFACE 
ERR_MISSING_CONFIG ERR_INVALID_CONFIG ERR_INVALID_NET_COMPRESSION 
ERR_INVALID_CONTEXT ERR_CANNOT_UNLOCK ERR_NO_CONTEXT_TABLE ERR_PUB_IP_IN_USE
ERR_PRIV_IP_IN_USE ERR_CANNOT_DEL_BRIDGE ERR_INVALID_IP ERR_NAT_FAILURE
ERR_IPFW_FAILURE ERR_ARP_FAILURE ERR_IFCONFIG_FAILURE ERR_INVALID_SYNTAX)],
netconst => [qw(ACK ACK_REQUEUE NAK_MALFORMED NAK_DUPLICATE NAK_INVALID 
NAK_STALE)], logging => [qw(nac_err nac_log)]);

Exporter::export_ok_tags('constants','logging','netconst');


#use Devel::NYTProf; # Profilier for debugging/optimization

###############################################################################
# Wrapper API
###############################################################################

#-----------------------------------------------------------
# Performs input sanity checking before calling nac_set_context
# with the requisite inputs
#
# Input: config ref, context, private ip, public ip, [client mac]
# Output: return code
#-----------------------------------------------------------
sub circe_set_ipfw_context {
	my ($confref,$context,$priv_ip,$pub_ip,$clientmac) = @_;

	unless ($clientmac) {
		$clientmac = "";
	}

	if (is_ipv4($priv_ip) && is_ipv4($pub_ip)) { 
		my ($ret) = nac_set_context($context,$priv_ip,$pub_ip,$clientmac,$confref);
		unless ($ret == 1) {
			nac_err("nac_set_context failed for $priv_ip with context $context");
		}
		return $ret;
	}
	return ERR_INVALID_IP;

}

#-----------------------------------------------------------
#-----------------------------------------------------------

sub circe_set_context {
    my ($confref,$context,$priv_ip,$pub_ip,$clientmac) = @_;


    my $attempt_limit = $confref->{"SET_CONTEXT_RETRY"};
    my $hostname = $confref->{"CONTEXT_DAEMON_HOST"};
    my $port = $confref->{"CONTEXT_DAEMON_PORT"};

    unless ($attempt_limit) { $attempt_limit = 1 }

    # Set up the socket
    LibNAC::nac_log("Creating socket for $hostname port $port","debug");
    my $context_daemon_sock = IO::Socket::INET->new (
        PeerAddr => $hostname,
        PeerPort => $port,
        Proto => 'tcp',
        Type => SOCK_STREAM,
		Blocking => 1,
    ) or die ("Failed to set up context daemon socket for $hostname:$port: $@");

    print $context_daemon_sock "$context $priv_ip $pub_ip $clientmac\r\n";

    my $response = <$context_daemon_sock>;
    my ($code,$message) = split(':',$response);

    if ($code == ACK_REQUEUE) {
        RETRY: for(my $i = 0; $i == $attempt_limit; $i++) {
            print $context_daemon_sock "$context, $priv_ip, $pub_ip, $clientmac";
            $response = <$context_daemon_sock>;
            ($code,$message) = split(':',$response);
            last RETRY if ($code != ACK_REQUEUE);
			sleep(1);
        }
    }

    if ($code != ACK && $code != ACK_REQUEUE) {
        # Something went wrong.
        LibNAC::nac_err("Error $code in circe-set-context: $message");
    } elsif ($code == ACK_REQUEUE) {
        LibNAC::nac_err("Context change $context $priv_ip $pub_ip failed "
            . "after $attempt_limit tries. Aborting attempt.");
    } elsif ($code == ACK) {
        # The check against the local session store succeeded. Now we wait
        # for the return code of the context change operation itself.
        $response = <$context_daemon_sock>;
        ($code,$message) = split (":",$response);
        unless ($code == 0) {
            LibNAC::nac_err("Error in circe-set-context: context change "
                . "$context $priv_ip $pub_ip failed with return $code");
        }
    }
    close $context_daemon_sock;
    return $code;
}

###############################################################################
# CONFIGURATION {{{1
###############################################################################

#-----------------------------------------------------------
# Reads the nac.conf file found at /usr/local/etc/nac.conf
# and returns a hash with options as keys.
#
# Input: none
# Output: Hash containing all options and their values
#-----------------------------------------------------------
sub nac_readconf {
	my ($config_path) = "/usr/local/etc/nac.conf";

	my $input = $_[$#_];
	if ($input && $input =~ m/^\//) {
		$config_path = $input;
	}	

	open( NACCONF, $config_path) or die("Error: can't find nac.conf");

	my(%confhash);
	my($line);
	my($nextline);

	while ($line = <NACCONF> ) {
		#ignore comments and empty lines
		unless ($line eq "\n" || $line =~ m/^#/ || !($line=~ m/=/)) {
			
			#Ensure that multi-line options are interpreted correctly
			if ($line=~ m/"/) {
				until ($line=~ m/[^=]"$/) {
					$nextline = <NACCONF>;
					$line = "$line" . "$nextline";
				}
			}
			elsif ($line=~ m/'/) {
				until ($line=~ m/[^=]'$/) {
					$nextline = <NACCONF>;
					$line = "$line" . "$nextline";
				}
			
			}
			#get rid of the trailing newline
			chomp($line);
			
			#Split the key and the value
			(my($key) , my($value))=split("=" , $line);
			
			#Strip quotation marks
			$value=~s["][]g;

			# Strip leading $. targeting table names 
			if ($value=~ m/\$/) {
				my @values = split(" ", $value);
				foreach my $entry (@values) {
					if ($entry =~ m/^\$/) {
						$entry =~ s/^\$//;
						$value=~ s/\$($entry)/$entry/;
					}
				}
			}	

			#print "key was $key, value was $value\n"; #debug code
			$confhash{"$key"}=$value;
		}
	}
	close NACCONF;
	return %confhash; 	
}

###############################################################################
# LOOKUP ASSOCIATED NET {{{2
###############################################################################

#-----------------------------------------------------------
# Looks to see what associated net is associated with the
# inputted id number.
#
# Input: an id number from ASSOCIATED_NETS map in nac.conf
# and a reference to the configration hash
# Output: two highest-order octets of an ip address
#------------------------------------------------------------
sub nac_associated_net_from_id {
	my ($id, $confref)		= @_;

	my @associated_nets = split(" ",$confref->{"ASSOCIATED_NETS"});

	# Iterate over associated nets, checking if the id number matches the one 
	# that was passed in.
	foreach my $net (@associated_nets) {
		if ($net =~ m/^($id)/) {
			$net =~ s/^($id);//;
			return $net;
		}
	}

	nac_err("WARNING:Invalid id $id in nac_associated_net_from_id:" .
				"Returning bogus ip and continuing");
	return '0.0';
}

###############################################################################
# LOOKUP HOSTNAME {{{2
###############################################################################

#-----------------------------------------------------------
# Find the current captive portal in a list of available
# portals mapped to their redirection urls.
#
# Input: reference to configuration hash
# Output: current captive portal address or hostname if
# address can't be found
#-----------------------------------------------------------
sub nac_default_url {
	my ($confref)  = @_;

	my $hostname = `/bin/hostname -s`;	
	chomp ($hostname);

	my $authhostmap = $confref->{"AUTH_HOST_MAP"};
	my @auth_host_maps = split(" ", $authhostmap);

	foreach my $hostmap (@auth_host_maps) {
		my ($host, $map) = split(":",$hostmap);
		if ($host eq $hostname) {
			($map)?return "http://$map/":return "http://" .
					`/bin/hostname -f` . '/';
		}
	}
}

###############################################################################
# LOOKUP INTERFACE {{{2
###############################################################################

#-----------------------------------------------------------
# Return interface associated for the given public IP
# address from the NAC_PUBLIC_NETS map.
#
# Input: public IP address
# Output: interface name
#-----------------------------------------------------------
sub nac_if_for_pub_ip {
	my ($pub_ip, $confref)	= @_;

	my $nac_pub_nets 	= $confref->{"NAC_PUBLIC_NETS"};
	my @nac_public_nets = split("\n",$nac_pub_nets);

	# Loop through all the public nets looking for the net
	# which the ip belongs to, then return the net's interface.
	foreach my $net (@nac_public_nets) {
		unless ($net eq "") {
			my ($subnet) = nac_subnet($net);
			$subnet =~ s/ *//;
			if (nac_ip_in_subnet($pub_ip,$subnet)) {
				return nac_if($net);
			}
		}
	}

	nac_err("Error in nac_if_for_pub_ip: No match found for ip $pub_ip");
	return ERR_NO_MATCHING_IFACE;
}

#-----------------------------------------------------------
# Return interface associated for the given private IP
# address via the NAC_PUBLIC_NETS map.
#
# Input: private IP address
# Output: interface name
#-----------------------------------------------------------
sub nac_if_for_pvt_ip {
	my ($priv_ip, $confref)	= @_;

	my $privatenets = $confref->{"NAC_PRIVATE_NETS"};
	my @nac_private_nets = split("\n", $privatenets);

	# Loop through all the private nets looking for the net
	# which the ip belongs to, then returns the nets interface.
	foreach my $net (@nac_private_nets) {
		my ($subnet) = nac_subnet($net);
		if (nac_ip_in_subnet($priv_ip,$subnet)) {
			return nac_if($net);
		}
	}

	nac_err("Error in nac_if_for_pvt_ip: No match found for ip $priv_ip");
	return ERR_NO_MATCHING_IFACE;
}

###############################################################################
# LOOKUP IPV6 BRIDGE {{{2
###############################################################################

#-----------------------------------------------------------
# Return the ipv6 bridge associated with a given ip address
#
# Input: Ip address and reference to configuration variable
# Output; bridge name
#-----------------------------------------------------------
sub nac_bridge_for_pvt_ip {
	my ($ip, $confref)	= @_;

	my $privatenets = $confref->{"NAC_PRIVATE_NETS"};
	my @nac_private_nets = split("\n",$privatenets);

	foreach my $netstr (@nac_private_nets) {
		my ($subnet) = nac_subnet($netstr);
		if (defined($subnet)) { 
			if (nac_ip_in_subnet($ip,$subnet)) {
				return nac_bridge($netstr);
			}
		}
	}
}

###############################################################################
# LOOKUP NATID {{{2
###############################################################################

#-----------------------------------------------------------
# Return a 7 digit unique identifier consisting of a netid
# and the last two octets of the client ip.
#
# Input: Public ip address for client
# Output: 7 digit identifier
#-----------------------------------------------------------
sub nac_nat_id {
	my ($public, $confref) = @_;

	my ($oct1,$oct2,$oct3,$oct4) = split('\\.',$public);

	# Pad the two octets so that the output will be a 7 digit identifier
	while (length($oct3) < 3) {
		$oct3 = "0" . "$oct3";
	}
	while (length($oct4) < 3) {
		$oct4 = "0" . "$oct4";
	}

	my ($netid)  = nac_associated_net_id($public,$confref);

	# This case is no longer accessible in a properly installed
	# circe implementation 
	# unless (defined($netid)) {
	#	 nac_err("Unable to acquire id number for given ip block $oct1.$oct2.");
	#	 $netid = 99;
	#}
	return "$netid$oct3$oct4";
}

###############################################################################
# LOOKUP NETID {{{2
###############################################################################

#-----------------------------------------------------------
# Return a netid based on the associated net that is
# associated with this particular public ip.
#
# Equivalent to nac_campus_net_id from libnac.sh. Change is primarily cosmetic
# Input: an IP number and a reference to a configuration hash
# Output: The ID number corresponding to the given net
#-----------------------------------------------------------
sub nac_associated_net_id {
	my ($pub_ip, $confref)	= @_;

	my @associated_nets	= split(" ", $confref->{"ASSOCIATED_NETS"});

	# Check if the input pub_ip begins with the first two octets in block.
	foreach my $net (@associated_nets) {
		my ($netid,$block) = split(";",$net);
		if ($pub_ip =~ m/^($block)/) {
			return $netid;
		}

	}

	nac_err("Invalid ip $pub_ip found no match within" .
		"known associated nets in nac_associated_net_id");
	nac_err("Check your nac.conf for inconsistencies or errors.");
	return 99;
}

###############################################################################
# LOOKUP NET STRINGS {{{2
###############################################################################

#-----------------------------------------------------------
# Parses a net-formatted string, such as NAC_PRIVATE_NETS or
# NAC_PUBLIC_NETS. Returns the value of the given "field".
#
# Input: a NET formatted string, and the field to return
# Output: returns the given field
#-----------------------------------------------------------
sub nac_net_parse {
	my ($net, $field)	= @_;

	$net =~ s/^\t+//;

	unless ($net) { return 0; }

	my ($subnet,$router,$clientrange,$if,$bridge) = split(";",$net);

	if ($field eq 'subnet') 		{ return $subnet; }
	elsif ($field eq 'router') 		{ return $router; }
	elsif ($field eq 'clientrange') { return $clientrange; }
	elsif ($field eq 'interface') 	{ return $if; }
	elsif ($field eq 'bridge') 		{ return $bridge; }
	elsif ($field eq 'all') {
		my %nethash = (
			'subnet' => $subnet,
			'router' => $router,
			'clientrange' => $clientrange,
			'interface' => $if,
			'bridge'	=> $bridge,
		);
		return %nethash;
	}
		
	nac_err("Invalid field $field called in nac_net_parse");
	return 0;
}

#-----------------------------------------------------------
# Returns the subnet field of a NET formatted string.
#
# Input: a NET formatted string
# Output: the subnet
#-----------------------------------------------------------
sub nac_subnet {
	my ($net) 	= @_;
	return nac_net_parse($net,"subnet");
}

#-----------------------------------------------------------
# Returns the router field of a NET formatted string.
#
# Input: a NET formatted string
# Output: the router
#-----------------------------------------------------------
sub nac_router {
	my ($net) 	= @_;
	return nac_net_parse($net,"router");
}

#-----------------------------------------------------------
# Returns the client ranges of a NET formatted string.
#
# Input: a NET formatted string
# Output: the client ranges
#-----------------------------------------------------------
sub nac_client_ranges {
	my ($net)	= @_;
	return nac_net_parse($net,"clientrange");
}

#-----------------------------------------------------------
# Returns the interface field of a NET formatted string.
#
# Input: a NET formatted string
# Output: the interface
#-----------------------------------------------------------
sub nac_if {
	my ($net)	= @_;
	return nac_net_parse($net,"interface");
}

#-----------------------------------------------------------
# Returns the bridge field of a NET formatted string.
#
# Input: a NET formatted string
# Output: the bridge
#-----------------------------------------------------------
sub nac_bridge {
	my ($net)	= @_;
	return nac_net_parse($net,"bridge");
}

###############################################################################
# CONTEXT {{{1
###############################################################################

#-----------------------------------------------------------
# Populate a hash table where n is the number of contexts
# requested.
#
# Input: references to an array of contexts, array of
# redirects.
# Output: hash linking contexts with their corresponding
# redirection pages.
#-----------------------------------------------------------
sub nac_context_populate {
	my ($contexts, $redirects)  = @_;
	my @contexts = @$contexts;

	my %context_hash;
	foreach my $i (0 .. $#contexts) {
		unless ($redirects->[$i] eq "none" ) {
			$context_hash{$contexts->[$i]} = $redirects->[$i];
		}
	}
	return %context_hash;
}

#-----------------------------------------------------------
# Returns the number of users in a given context.
# 
# Input: context name (or "all"), and reference to
# configuration file
# Output: Number of users in given context
# ----------------------------------------------------------
sub nac_auth_count {
	my ($context, $confref) 	= @_;
	my $count = 0;

	unless ($context && $confref) {
		nac_err("no context specified in nac_auth_count");
		return ERR_INVALID_CONTEXT;
	}

	if ($context eq "all") {
		$context = $confref->{"CONTEXTS"}; 
	}
	my @contexts = split(" ",$context);

	foreach my $context (@contexts) {
		my @context_list = nac_list_context($context,$confref);
		$count = $count + $#context_list + 1;
	}
	return $count;
}

###############################################################################
# GET CONTEXT {{{2
###############################################################################

#-----------------------------------------------------------
# Gets the context that this ip address is recorded as 
# having.
#
# Input: IP address, reference to configuration hash
# Output: context name 
#-----------------------------------------------------------
sub nac_get_context {
	my ($ip, $confref) = @_;

	my @public_tables 	= split(' ', $confref->{"PUBLIC_TABLES"});
	my @private_tables 	= split(' ', $confref->{"PRIVATE_TABLES"});
	my @tables = (@private_tables, @public_tables);

	# Loop through all the tables looking for the one the ip
	# address belongs to, if found return its context.
	foreach my $table (@tables) {
		my ($firewall_tables) = $confref->{"$table"};
		if (nac_table_check($ip, $firewall_tables))  {
			return nac_table_to_context($firewall_tables, $confref);
		}
	}
	my ($DEFAULT_CONTEXT) = $confref->{"DEFAULT_CONTEXT"};
	nac_log("No match found for ip address $ip," . 
				"returning default context $DEFAULT_CONTEXT","err");
	return $DEFAULT_CONTEXT;
}

###############################################################################
# LOOKUP NAT {{{2
###############################################################################

#-----------------------------------------------------------
# Print and return a human readable list of nat rules for a
# given context.
#
# Input: Context name and reference to configuration file
# Output: Array list of ip's and nac id's
#-----------------------------------------------------------
sub nac_list_context {
	my ($context, $confref) = @_;
	
	my $table = nac_context_to_priv_table($context,$confref);
	my @context_list;
	if ($table == 0) { return @context_list;} 

	my $table_entries = nac_dump_table($table);
	
	my @table_entries = split("\n",$table_entries);

	# Grab an ip from the table, strip out the CIDR notation netmask,
	# print it with the nac_ip and push them to the context_list
	foreach my $entry (@table_entries) {
		my ($ip, $natid)= split(" ",$entry);
		$ip =~ s/\/[0-9]*//;
		my $nac_ip = nac_ip_from_onat_id($natid,$confref);
		my $combined_ips = "$ip\t$nac_ip";
		print "$combined_ips\n";
		push(@context_list,"$combined_ips");
	}
	return @context_list;
}

###############################################################################
# LOOKUP TABLE {{{2
###############################################################################

#-----------------------------------------------------------
# Print the private table number associated with a given
# context.
#
# Input: context name and reference to configuration hash
# Output: private table number
#-----------------------------------------------------------
sub nac_context_to_priv_table {
	my ($context, $confref)  = @_;

	if ($confref->{uc("$context")}){
		my ($table, undef)	= split(" ",$confref->{uc("$context")});
		return $table;
	} else {
		return 0;
	}
}

#-----------------------------------------------------------
# Print the public table number associated with a given
# context.
#
# Input: Context name and reference to configuration hash
# Output: public table number
#-----------------------------------------------------------
sub nac_context_to_pub_table {
	my ($context, $confref)  = @_;

	my (undef, $table)	= split(" ",$confref->{uc("$context")});
	return $table;
}

###############################################################################
# LOOKUP TABLE CONTEXT {{{2
###############################################################################

#-----------------------------------------------------------
# Converts a table number to its associated context by
# checking it against a list stored in nac.conf. Uses the
# whitespace-separated contexts list as well.
#
# Input: table to convert, reference to configuration hash
# Output: Context name
#-----------------------------------------------------------
sub nac_table_to_context {
	my ($table, $confref)    = @_;

	my $contexts = $confref->{"CONTEXTS"};
	my @contexts = split(' ', $contexts);

	# For each context in the context statement, check the table mapping
	# and return the context if a match is found
	foreach my $context (@contexts) {
		if (exists $confref->{uc("$context")}) {
			my ($contexttables) = $confref->{uc("$context")};
			if (index($contexttables, $table) >= 0) {
				return $context;
			}
		}
	}
}    	

###############################################################################
# SET CONTEXT {{{2
###############################################################################

#-----------------------------------------------------------
# Sets the context of the given client to whatever context
# is given as an argument. 
#
# Input: New client context, client private and public ip,
# client mac, and reference to configuration hash.
# Output: Boolean recording success.
#-----------------------------------------------------------
sub nac_set_context {
	my ($context, $priv_ip, $pub_ip, $clientmac, $confref) = @_;
	my $ret;

	my @contexts = $confref->{"CONTEXTS"};

	unless ($clientmac) { 
		$clientmac="";
	}

	# If new context is unauth disable nat mapping for client.
	if ($context eq "unauth") {
		nac_log("Context Change: $context $priv_ip $pub_ip $clientmac","info");
		$ret = nac_disable_nat($priv_ip,$pub_ip,$clientmac,$confref);
	}
	# Else establish nat mapping with appropriate context.
	elsif ($confref->{uc($context)}) {
		nac_log("Context Change: $context $priv_ip $pub_ip $clientmac","info");
		$ret = nac_enable_nat($priv_ip,$pub_ip, $context,$clientmac,$confref);
	} 
	else {
		nac_err("Error in nac_set_context: context $context not matched");
		$ret = ERR_INVALID_CONTEXT;
	}

	return $ret;
}

###############################################################################
# FLUSH STATE {{{1
###############################################################################

#-----------------------------------------------------------
# Flush the state of all clients- This is a dangerous and
# irreversible operation that will terminate all active
# clients and place everyone in the unauth context.
#
# Input: Reference to configuration hash
# Output: None
#-----------------------------------------------------------
sub nac_flush_state {
	my ($confref)	= @_;

	my @contexts	= split(" ",$confref->{"CONTEXTS"});
	my @priv;
	my @pub;
	my @clientmac;
	my @context_list;

	# Build arrays containing private, public, and mac addresses
	foreach my $context (@contexts) {
		@context_list	= nac_list_context($context,$confref);
		foreach my $entry (@context_list) {
			my($priv, $pub) = split("\t",$entry);
			my($arp)	= `/usr/sbin/arp $priv`;
			my ( undef, undef, undef, $clientmac, undef) = split(" ",$arp);
			push(@priv,$priv);
			push(@pub,$pub);
			push(@clientmac,$clientmac);
		}
	}

	# Disable the netgraph nat for found in the context tables.
	for (my $i = 0; $i <= $#priv; $i++) {
		nac_disable_nat($priv[$i],$pub[$i],$clientmac[$i],$confref);
	}

	my @public_tables	= split(" ",$confref->{"PUBLIC_TABLES"});
	my @private_tables	= split(" ",$confref->{"PRIVATE_TABLES"});
	my @tables = (@public_tables, @private_tables);
	my ($tableno);

	# Remove the ipfw rules for all the private and public tables.
	foreach my $table (@tables) {
		$tableno = $confref->{"$table"};
		unless(nac_ipfw("table $tableno flush") == 1) {
			nac_err("Error flushing table $table in nac_flush_state");
		}
	}

	if (open(PIDFILE,"/var/run/circe-daemon.pid")) {
		print "Daemon is running, restarting.\n";
		my $pid = <PIDFILE>;
		chomp($pid);
		if (nac_is_running($pid)) {
			`/usr/local/etc/rc.d/circed restart`;
		}
	}
}

###############################################################################
# RULE TABLES {{{2
###############################################################################

#-----------------------------------------------------------
# Returns a scalar containing the contents of the given
# ipfw table.
#
# Input: Table number
# Output: scalar with table entries (newline seperated)
#-----------------------------------------------------------
sub nac_dump_table {
	my ($table)		= @_;

	my $table_entries 	= `/sbin/ipfw table $table list`;
	return $table_entries;
}

#-----------------------------------------------------------
# Function that checks for a given ip in a ipfw table.
# 
# Input: ip address and table
# Output: 0 if ip address is not found, 1 if it is.
#-----------------------------------------------------------
sub nac_table_check {
	my ($ip, $table) = @_;

	my @table_entries = split("\n",`/sbin/ipfw table $table list`);

	foreach my $entry (@table_entries) {
		(my($table_ip),undef) = split(' ', $entry);
		if ($table_ip=~m/^$ip\//) {
			return 1;
		}
	}
	return 0;
}  	

###############################################################################
# IPV6 {{{1
###############################################################################

#-----------------------------------------------------------
# Add ipv6 connectivity for the given client via the bridge
# interface.
#
# Input: Client MAC, Bridge interface, source interface
# Output: boolean of subroutine success
#-----------------------------------------------------------
sub nac_add_v6 {
	my ($clientmac, $bridgeif, $sourceif) = @_;

	my $ret = nac_ifconfig("$bridgeif static $sourceif $clientmac");

	unless ( $ret == 1 ) {
		nac_err("ifconfig bridge creation failed for client mac" .
					"$clientmac in nac_add_v6");
		return $ret;
	}
	return 1;
}

#-----------------------------------------------------------
# Delete a given mac address on the given ipfw map.
#
#Input: Client mac and bridge interface
#Output: Boolean representing subroutine success
#-----------------------------------------------------------
sub nac_del_v6 {
	my ($clientmac, $bridgeif)	= @_;

	unless ( nac_ifconfig("$bridgeif deladdr $clientmac") == 1) {
		nac_err("ifconfig address delete failed with interface $bridgeif" .
					"and address $clientmac in nac_del_v6");
		return ERR_CANNOT_DEL_BRIDGE;
	}
	return 1;
}

###############################################################################
# LOGGING {{{2
###############################################################################

#-----------------------------------------------------------
# Simple logger. logs to the local2 facility with the tag
# "nac". Priority is set per-call.
#
# Inputs: message to log and priority to log at
# Output: return code 1
#
# TODO: Have priority to log configurable in nac.conf
# TODO; externalize setlogsock, openlog, and closelog to occur only once per 
# session
# TODO: Make portable to non-unixes?
#-----------------------------------------------------------

sub nac_log {
	my ($line, $priority)  	= @_;

	setlogsock('unix');
	openlog("nac",'','local2');
	syslog($priority, $line);
	closelog;
	return 1;
}

#-----------------------------------------------------------
# Error logger for circe scripts, logs by calling nac_log
# with "err" priority.
#
# Input: line to log
# Output: return code 1
#-----------------------------------------------------------

sub nac_err {
	my ($line) 		= @_;
	nac_log($line,"err");
	return 1;
}

###############################################################################
# LOCKFILE {{{1
###############################################################################

#-----------------------------------------------------------
# Locks a given lockfile and writes to the list of running locks
#
# Input: Lockfile to lock, identifier for lock, reference
# 		 to configuration hash
# Output: Boolean of lock succcess
#-----------------------------------------------------------

sub nac_lock {
	my ($lockfile, $ident, $confref) = @_;

	# Get information from configuration hash
	my $time_max		= $confref->{"LOCK_SLEEP_LIMIT"};
	
	# Start a timer to keep track of elapsed time
	my $start			= time();
	my $elapsedtime		= 0;

	# Execute loop until the lockfile no longer exists or a return statement 
	# is reached
	while ( -e $lockfile ) {
		my $runningpid = `/bin/cat $lockfile`;

		# If the pid denoted in the lockfile is still running, wait and
		# decrement the counter; otherwise remove the lockfile
		if (nac_is_running($runningpid)) {
			nac_log("$ident waiting for lock $lockfile" .
						"held by pid $runningpid","info");
			sleep(1);
			$elapsedtime = time() - $start;
			if ($elapsedtime >= $time_max) {
				nac_err("Error: $ident could not obtain" .
							"lock $lockfile held by $runningpid");
				return ERR_CANNOT_UNLOCK;
			}
		} else {
			nac_log("$ident removing lock $lockfile" .
						"formerly held by $runningpid","info");
			unlink($lockfile);
		}
	}
	# Creates the lockfile and prints the current pid
	open(LOCK,">$lockfile") || die "Error: Failed to open lockfile $lockfile.";
	print LOCK "$$";
	close LOCK;

	return 1;
}

#-----------------------------------------------------------
# Unlock the given lock file
#
# Input: Lockfile to unlock, reference to configuration hash
# Returns 1.
#-----------------------------------------------------------
sub nac_unlock {
	my ($lockfile, $confref)	= @_;

	my @newlocks;

	nac_log("Removing lockfile $lockfile in nac_unlock.","info");

	# Remove the lockfile
	unlink($lockfile);

	return 1;
}


###############################################################################
# NAT {{{1
###############################################################################
# All of the functions that are associated with Netgraph Nat Nodes.

##############################################################################
# LOOKUP FUNCTIONS {{{2
##############################################################################

#-----------------------------------------------------------
# Return the name of the ng_nat node associated with a
# given public ip.
#
# Inputs: The public ip address and a reference to a
# configuration hash
# Outputs: netgraph name id
#-----------------------------------------------------------
sub nac_nat_name {
	my ($pub_ip, $confref)  = @_;

	my $name_prefix		= $confref->{"NAME_PREFIX"};
	my $natid			= nac_nat_id($pub_ip,$confref);

	return "$name_prefix$natid"; 
}

#-----------------------------------------------------------
# Return a netgraph cookie value for inbound NAT
#
# Inputs: the public ip address and a reference to a
# configuration variable
# Output: netgraph cookie id
#-----------------------------------------------------------
sub nac_inat_id {
	my ($pub_ip, $confref) = @_;

	my $prefix	 = $confref->{"INATPREFIX"};
	my $natid	 = nac_nat_id($pub_ip,$confref);

	return "$prefix$natid";
}

#-----------------------------------------------------------
# Return a netgraph cookie value for outbound NAT
#
# Inputs: the public ip address and a reference to a
# configuration hash
# Output: netgraph cookie id
#-----------------------------------------------------------
sub nac_onat_id {
	my ($pub_ip, $confref) = @_;

	my $prefix 	= $confref->{"ONATPREFIX"};
	my $natid	= nac_nat_id($pub_ip,$confref);

	return "$prefix$natid";
}

###############################################################################
# REVERSE LOOKUP FUNCTIONS {{{2
###############################################################################

#-----------------------------------------------------------
# Prints a public IP address associated with a given
# netgraph cookie 
#
# Input: a netgraph cookie id and a reference to a
# configuration hash
# Output: Fully-formed ip address from the cookie
#-----------------------------------------------------------
sub nac_ip_from_onat_id {
	my ($natid, $confref) 	= @_;

	# Determine the length of the nat prefix and eliminate it
	my $prefixlen 	= length($confref->{"ONATPREFIX"});
	my $natidlen	= length($natid);
	$natid 			= substr($natid,$prefixlen,$natidlen-$prefixlen);	

	# Retrieve a netid and call nac_associated_net_from_id to get the first
	# two ip octets
	my $assoc_nets 	= $confref->{"ASSOCIATED_NETS"};
	my @nets		= split(";",$assoc_nets);
	my $net_len 	= length($nets[0]);


	my $netid		= substr($natid,0,$net_len);
	my $net			= nac_associated_net_from_id($netid,$confref);

	# The third and fourth octets come from the input string
	$natid			= substr($natid,$net_len,$natidlen-$prefixlen-$net_len);
	my $oct3 		= substr($natid,0,3);
	my $oct4 		= substr($natid,3,3);

	# Strip any unnecessary leading 0's 
	$oct3 =~ s/^0*//;
	$oct4 =~ s/^0*//;

	# Return fully-formed ip address
	return "$net.$oct3.$oct4";
}

###############################################################################
# NODE CREATION {{{2
###############################################################################

#-----------------------------------------------------------
# Create a new netgraph nat node for the given client.
# 
# Input: Private IP, Public IP, reference to config hash
# Output: Returns 1 (or dies)
# ----------------------------------------------------------
sub nac_new_ng_nat {
	my ($priv_ip, $pub_ip, $confref) = @_;

	my $onatid		= nac_inat_id($pub_ip,$confref);
	my $inatid		= nac_onat_id($pub_ip,$confref);
	my $name		= nac_nat_name($pub_ip,$confref);

	unless ($onatid !~ m/^[0-9]99/ && $inatid !~ m/^[0-9]99/ && defined($name)){
		nac_err("Malformed incoming or outgoing natid in nac_new_ng_nat.");
		return ERR_INVALID_NET_COMPRESSION;
	}

	try {
		if (nac_ngctl("mkpeer ipfw: nat $onatid out") == ERR_NAT_FAILURE)
			{ die ("ngctl failed to make outgoing nat") }
		if (nac_ngctl("name ipfw:$onatid $name") == ERR_NAT_FAILURE)
			{ die ("ngctl failed to set name $name to node") }
		if (nac_ngctl("connect ipfw: $name: $inatid in") == ERR_NAT_FAILURE)
			{ die ("ngctl failed to connect nodes") }
		if (nac_ngctl("msg $name: setaliasaddr $pub_ip") == ERR_NAT_FAILURE)
			{ die ("ngctl failed to setaliasaddr") }
		if (nac_ngctl("msg $name: redirectaddr " . '{local_addr='. 
					"$priv_ip" . ' alias_addr=' . "$pub_ip" .  '}')
					== ERR_NAT_FAILURE) 
			{ die ("ngctl failed to set redirectaddr") }
		}
	catch {
		nac_err("Error in nac_new_ng_nat:" .
					"Failed to create new node for $priv_ip $pub_ip");
		return ERR_NAT_FAILURE;
	}
	return 1;
}

#-----------------------------------------------------------
# Create a new in-kernel nat node for the given client.
# 
# Input: Private IP, Public IP, reference to config hash
# Output: Return Code
# ----------------------------------------------------------
sub nac_new_ipfw_nat {
	my ($priv_ip, $pub_ip, $confref) = @_;

	my $natid = nac_inat_id($pub_ip,$confref);

	if ($natid =~ m/^[0-9]99/) {
		nac_err("Error in nac_new_ipfw_nat: " .
			"Malformed natid $natid.");
		return ERR_INVALID_NET_COMPRESSION;
	}

	try {
		unless (nac_ipfw("nat $natid config redirect_addr $priv_ip $pub_ip") == 1) {
			die;
		}
	} catch {
		nac_err("Error in nac_new_ipfw_nat: " .
			"Failed to create new ipfw nat $natid for $priv_ip $pub_ip");
		return ERR_NAT_FAILURE;
	}
	return 1;
}




#-----------------------------------------------------------
# Bring up a new NAT node for the given private ip to the
# given public ip.
#
# Inputs: private and public ip, private and public table
# number, context, client mac, reference to configuration
# hash.
# Output: Return code 
#-----------------------------------------------------------
sub nac_enable_nat {
	my ($private_ip, $public_ip, $context, $clientmac, $confref) = @_;

	my $lockfile		= $confref->{"LOCKFILE"};
	my $natid;
	my $onatid;
	my $inatid;

	# Initialize all needed variables from inputs.
	my $private_table	= nac_context_to_priv_table($context,$confref);
	my $public_table	= nac_context_to_pub_table($context,$confref);

	# Determine if the user wants to use IPFW NAT or Netgraph
	if ($confref->{"USE_KERNEL_NAT"}) {
		# Only one natid needed now.
		$natid			= nac_inat_id($public_ip,$confref);
		$onatid 		= $natid;
		$inatid 		= $natid;
	} else { 
		# Netgraph case; seperate inat and onat ids
		$onatid 		= nac_inat_id($public_ip, $confref);
		$inatid			= nac_onat_id($public_ip, $confref);
	}

	my $pubif			= nac_if_for_pub_ip($public_ip,$confref);
	my $pubif_mac		= nac_if_mac($pubif);
	my $sourceif		= nac_if_for_pvt_ip($private_ip,$confref);
	my $bridgeif		= nac_bridge_for_pvt_ip($private_ip,$confref);

	my $ret;

	if ($onatid  =~ m/^[0-9]99/ || $inatid =~ m/^[0-9]99/) {
		nac_err("Malformed natid in nac_enable_nat.");
		nac_err("Check the ASSOCIATED_NETS map in nac.conf" .
					"for inconsistencies.");
		nac_err("Aborting context change operation.");
		return ERR_INVALID_NET_COMPRESSION;
	}

	# Make sure you can obtain the lockfile for both the public
	# and private ip addresses.
	unless (nac_lock("$lockfile-$public_ip",$clientmac,$confref)) {
		nac_err("Could not obtain lockfile in nac_enable_nat" .
					"for client $private_ip $clientmac; aborting");
		return ERR_CANNOT_UNLOCK;
	}
	if (nac_lock("$lockfile-$private_ip",$clientmac,$confref) 
					== ERR_CANNOT_UNLOCK) {
		nac_err("Could not obtain lockfile in nac_enable_nat" .
					"for client $private_ip $clientmac; aborting");
		return ERR_CANNOT_UNLOCK;
	}

	# Make sure that the ip addresses link to someone in the
	# unauth state.
	unless (nac_get_context($public_ip,$confref) eq "unauth") {
		my $get_context = nac_get_context($public_ip,$confref);
		nac_err("Nac_get_context returned $get_context"); 
		nac_err("Public IP $public_ip already in use.");
		return ERR_PUB_IP_IN_USE;
	}
	unless (nac_get_context($private_ip,$confref) eq "unauth") {
		nac_err("Private IP $private_ip already in use.");
		return ERR_PRIV_IP_IN_USE;
	}

	# Try to create the netgraph nat node for the address, then
	# do a series of arp requests, and add the ip addresses to
	# the appropriate ipfw tables. If it fails disable the nat
	# nodes.
	try {
		if ($confref->{"USE_KERNEL_NAT"}) {
			unless (nac_new_ipfw_nat($private_ip,$public_ip,$confref) == 1) {
				die;
			}
		} else {
			unless (nac_new_ng_nat($private_ip,$public_ip,$confref) == 1) {
				die;
			}
		}
		nac_arp("-d $public_ip");
		if (nac_arp("-s $public_ip $pubif_mac pub") == 23) {
			die;
		}
		unless (nac_ipfw("table $private_table add $private_ip $onatid") == 1){
			die;
		}
		unless (nac_ipfw("table $public_table add $public_ip $inatid") == 1) {
			die;
		}
		if ($confref->{"USE_V6"} && $context eq "registered") {
			nac_add_v6($clientmac,$bridgeif,$sourceif) 
				|| die ("Can't create v6 bridge for $private_ip" .
							"in nac_enable_nat");
		}
	}
	catch {
		nac_err("Caught exception in nac_enable_nat: Cleaning up");
		$ret = $@;
		nac_disable_nat($private_ip,$public_ip,$clientmac,$confref);
		return $ret;
	}

	nac_log("nac_enable_nat: Left critical section","info");

	# Release the lock files.
	nac_unlock("$lockfile-$public_ip",$confref);
	nac_unlock("$lockfile-$private_ip",$confref);
	return 1;
}

###############################################################################
# NODE DESTRUCTION {{{2
###############################################################################

#-----------------------------------------------------------
# Destroy the indicated in-kernel nat node for the given client.
# 
# Input: Public IP, reference to config hash
# Output: Returns 1 (or dies)
# ----------------------------------------------------------
sub nac_destroy_ipfw_nat {
	my ($pub_ip,$confref) = @_;
	my $natid = nac_inat_id($pub_ip,$confref);

	if ($natid =~ m/^[0-9]99/) {
		nac_err("Error in nac_destroy_ipfw_nat: " .
			"Malformed natid $natid.");
		return ERR_INVALID_NET_COMPRESSION;
	}
		
	try {
		unless (nac_ipfw("nat delete $natid") == 1) {
			die;
		}
	} catch {
		nac_err("Error in nac_destroy_ipfw_nat: " .
			"Failed to destroy ipfw nat $natid for $pub_ip. Recommend " .
			"manually deleting this node.");
		return ERR_NAT_FAILURE;
	}
	return 1;
}

#-----------------------------------------------------------
# Destroy a ng_nat node for a given public IP address.
#
# Input: Public ip
# Output: boolean subroutine status 
#-----------------------------------------------------------
sub nac_destroy_ng_nat {
	my ($public_ip, $confref) = @_;

	my $onatid = nac_inat_id($public_ip,$confref);

	if ($onatid =~ m/^[0-9]99/) {
		nac_err("Failed to receive valid inat id in nac_destroy_ng_nat");
		return ERR_INVALID_NET_COMPRESSION;
	}

	my $ret = nac_ngctl("shutdown ipfw:$onatid");

	unless ($ret == 1)  {
		nac_err("ngctl shutdown failed with id $onatid and exit code $?");
		return ERR_NAT_FAILURE;
	}
	return 1;
}

#-----------------------------------------------------------
# Disable ng nat described by the given network addresses
#
# Input: private and public ip, mac address, and reference
# to configuration hash.
# Output: Returns 1
#-----------------------------------------------------------
sub nac_disable_nat {
	my ($priv_ip, $pub_ip, $client_mac, $confref)	= @_;

	my $inatid;
	my $onatid;
	my $natid;

	# Acquire a natid. How we populate these variables depends on the 
	# nat implementation used
	if ($confref->{"USE_KERNEL_NAT"}) {
		# We only need one natid now with ipfw nat
		$natid		= nac_inat_id($pub_ip,$confref);
		# Alias onatid and inatid to natid to simplify checks later
		$inatid 	= $natid;
		$onatid		= $natid;
	} else {
		# Directionality is required in the netgraph case
		$onatid      = nac_inat_id($pub_ip,$confref);
		$inatid      = nac_onat_id($pub_ip,$confref);
	}

	my $bridge_if	= nac_bridge_for_pvt_ip($priv_ip,$confref);

	if ($onatid =~ m/^[0-9]99/ || $inatid =~ m/^[0-9]99/) {
		nac_err("Malformed nat id received in nac_disable_nat.");
		nac_err("Check your nac.conf for errors in your ASSOCIATED_NETS map.");
		nac_err("Aborting context change operation.");
		return ERR_INVALID_NET_COMPRESSION;
	}	

	# Destroy the arp entry and the nat entry
	nac_arp("-d $pub_ip");

	my $ret; 

	if ($confref->{"USE_KERNEL_NAT"}) {
		$ret 	= nac_destroy_ipfw_nat($pub_ip,$confref);
	} else {
		$ret 	= nac_destroy_ng_nat($pub_ip,$confref);
	}

	unless ( $ret  == 1) {
		nac_err("Error in destroying nat instance $priv_ip $pub_ip with
			error code $ret");
		return $ret;
	}

	# If using ipv6 remove it from the ipv6 bridge.
	if ($confref->{"USE_V6"}) {
		$ret = nac_del_v6($client_mac,$bridge_if);
		unless ($ret == 1) { 
			nac_err("Failed to delete bridge for ipv6 operation.");
			return $ret;
		}
	}

	my @private_tables	= split(" ",$confref->{"PRIVATE_TABLES"});
	my @public_tables	= split(" ",$confref->{"PUBLIC_TABLES"});
	
	my $tableno;

	# Destroy the ipfw table entries for this nat node.
	foreach my $table (@private_tables) {
		$tableno = $confref->{"$table"};
		nac_ipfw("-q table $tableno delete $priv_ip $onatid");
	}
	foreach my $table (@public_tables) {
		$tableno = $confref->{"$table"};
		nac_ipfw("-q table $tableno delete $pub_ip $inatid");
	}
	return 1;
}

###############################################################################
# REDIRECT {{{1
###############################################################################

#-----------------------------------------------------------
# Based on the client-request URL redirect the client to
# the appropriate place depending on whetheror not the
# URL requested is allowed.
#
# Input: A line containing the requested url and the ip of
# the client, and a reference to the configuration hash.
# Output: Blank line is no redirections, or a redirect
# statement for squid to pass to client.
#-----------------------------------------------------------
sub nac_redirect {
	my ($line, $confref,$context_linked_hashref) = @_;
	# $context_linked_hashref and $cipher may not be defined, depending on
	# the values of CONTEXTUAL_REDIRECT and REDIRECT_ENCRYPT_URL, respectively

	#nac_log("Line: $line","debug");
	chomp($line);

	my ($req_url, $private_ip,undef,undef,$remote_ip) 	= split(/[\s]+/, $line);

	$private_ip				=~ s/\/-$//;
	$remote_ip				=~ s/^myip\=//;

	my $req_url_lc 				= lc($req_url);
	my $redirect;

	my $authorized_urls 		= $confref->{"AUTHORIZED_URLS"};
	my $table_https_hosts		= $confref->{"TABLE_CIRCE_HTTPS_HOSTS"};
	my @auth_urls  				= split(/[\s]+/, $authorized_urls);
	chomp(@auth_urls);
	my $redirect_url			= nac_default_url($confref);

	# If the request is for an authorized URL let it through.
	foreach my $url (@auth_urls) {
		#nac_log("Checking authorized url $url in nac_redirect","debug");
		my $url_lc = lc($url);
		if ($req_url_lc =~ /($url_lc)/ && $url) {
	# Commented out redirect logging for privacy purposes.
	# TODO: Make this configurable
	#		nac_log("Match found against authorized url" . 
	#					" $url_lc in LibNAC::nac_redirect","info");
			return;
		}	
	}

	# If the config file specifies the CONTEXTUAL_REDIRECT flag, we 
	# redirect the user based on which context they reside in. Otherwise,
	# just return the default redirection url
	if ($confref->{"CONTEXTUAL_REDIRECT"}) {
		my $client_context	= nac_get_context($private_ip,$confref);
		unless ($context_linked_hashref->{$client_context} eq "none" ) {
			($redirect) = sprintf("302:%s%s", "$redirect_url",
							"$context_linked_hashref->{$client_context}");
		} else {
			($redirect) = sprintf("302:%s", "$redirect_url");
		}

	} 
	else {
		($redirect) = sprintf("302:%s","$redirect_url");
	}
	#nac_log("Redirect completed for $private_ip from $req_url" .
	#			"to $redirect","info");
	
	if ($confref->{"REDIRECT_INCLUDE_URL"})  {
		if ($confref->{"REDIRECT_ENCRYPT_URL"}) {
			my $key;
			my $do_not_encrypt=0;
			if ($confref->{"REDIRECT_ENCRYPT_KEY"}) {
				$key = $confref->{"REDIRECT_ENCRYPT_KEY"};
			} elsif ($confref->{"REDIRECT_ENCRYPT_KEYFILE"}){
				my $keyfile = $confref->{"REDIRECT_ENCRYPT_KEYFILE"};
				open(KEYFILE,"$keyfile");
				$key = <KEYFILE>;
				chomp($key);
			} else {
				nac_err("No suitable encryption key found, not encrypting");
				$do_not_encrypt=1;
			}

			unless ($do_not_encrypt) {
				my $cipher = Crypt::CBC->new(
						-key => $key,
						-cipher => "Crypt::OpenSSL::AES"
				);

				$req_url = $cipher->encrypt($req_url);
				$req_url = unpack("H*",$req_url);
			}
		}
		return sprintf("%s%s", "$redirect" , "?$req_url");
	} 
	else {
		return $redirect;
	}
}

###############################################################################
# UTIL {{{1
###############################################################################

#-----------------------------------------------------------
# Checks to see if the given process is running by using ps
# against the given pid and checking the output. Process is 
# running if ps returns 2 lines (one of which is the
# header).
#
# Input: PID of process to check
# Output: Boolean of pid status
#-----------------------------------------------------------

sub nac_is_running {
	my ($pid)		= @_;

	my $pslines		= `/bin/ps $pid | /usr/bin/wc -l`;
	$pslines		-= 1;
	return $pslines;
}

#-----------------------------------------------------------
# Maps ip address to integer equivalent.
#
# Input: Ip address
# Output: decimal equivalent
#-----------------------------------------------------------
sub nac_ip_to_num {
	my ($ip)	= @_;

	my ($oct1,$oct2,$oct3,$oct4) = split(/\./,$ip);
	my $num 	= $oct1*2**24 + $oct2*2**16 + $oct3*2**8 + $oct4;
	return $num;
}

#-----------------------------------------------------------
# Returns the mac address for a given interface.
# 
# Input: Interface
# Output: mac address
#-----------------------------------------------------------
sub nac_if_mac {
	my ($iface)		= @_;

	my $iface_info		= `/sbin/ifconfig $iface`;
	my @iface_data		= split("\n",$iface_info);
	my $hwaddr;

	# Loop through all ifconfig data and 
	foreach my $line (@iface_data) {
		if ($line =~ m/^[\s]+ether/) {
			(undef,$hwaddr)	= split(" ",$line);
		}
	}
	return $hwaddr;
}

#-----------------------------------------------------------
# Checks if a given ip address exists in the given subnet
# (CIDR notation).
#
# Input: Ip address and subnet
# Output: Boolean 
#-----------------------------------------------------------
sub nac_ip_in_subnet {
	my ($ip, $subnet)	= @_;

	return Net::IP::Match::XS::match_ip($ip, $subnet);	
}

###############################################################################
# COMMAND LINE WRAPPERS {{{2
###############################################################################

#-----------------------------------------------------------
# Wrapper for system call to arp, logs to nac_log if the 
# call fails.
#
#Input: Arguments for arp
#Output: Boolean of arp success
#-----------------------------------------------------------
sub nac_arp {
	my ($args)	= @_;
	my $ret = nac_system("/usr/sbin/arp $args");
	unless ($ret == 0) {
		nac_err("Error in nac_arp: arp returned code $ret with args $args");
		return ERR_ARP_FAILURE;
	}
	return 1;
}

#-----------------------------------------------------------
# Wrapper for system call to ifconfig, returns 0 if the
# call fails and logs to nac_log.
#
# Input: argument line for ifconfig
# Output: boolean of ifconfig success
#-----------------------------------------------------------
sub nac_ifconfig {
	my ($args)	= @_;
	my $ret = nac_system("/sbin/ifconfig $args");
	unless ($ret == 0) {
		nac_err("Error in nac_ifconfig: ifconfig returned" .
					"code $ret with args $args");
		return ERR_IFCONFIG_FAILURE;
	}
	return 1;
}

#-----------------------------------------------------------
# Wrapper for system call to ipfw, returns 0 if the call
# fails and logs to nac_log.
#
# Input: Argument line for ipfw
# Output: boolean of ipfw success
#-----------------------------------------------------------
sub nac_ipfw {
	my ($args)	= @_;
	my $ret = nac_system("/sbin/ipfw $args");
	unless ($ret == 0) {
		nac_err("Error in nac_ipfw: ipfw returned code $ret with args $args");
		return ERR_IPFW_FAILURE;
	}
	return 1;
}

#-----------------------------------------------------------
# Wrapper for system call to ngctl, returns 0 if the call
# fails and logs to nac_log.
#
# Input: argument line for ngctl
# Output: boolean of ngctl success
#-----------------------------------------------------------
sub nac_ngctl {
	my ($args)	= @_;
	my $ret = nac_system("/usr/sbin/ngctl $args");
	unless ($ret == 0) {
		nac_err("Error in nac_ngctl: ngctl returned code $ret with args $args");
		return ERR_NAT_FAILURE;
	}
	return 1;
}

#-----------------------------------------------------------
# Wrapper for the system() function call that redirects STDOUT
# and STDERR and sets the local SIGCHLD handler to null for
# daemon compatibility reasons.
#-----------------------------------------------------------
sub nac_system {
	local $SIG{CHLD} = '';
	my $cmd = shift;
	system("$cmd 1> /dev/null");
	my $ret = ($? >> 8);
	return $ret;
}

###############################################################################
# vim:ft=perl:ts=4:foldmethod=marker
1;
